#!/usr/bin/perl

########################################################
#
# ncDNAlign.3.mergeGwAln.pl
#    part of the NcDNAlign alignment pipeline, step (3)
#
# Authors: Dominic Rose, Kristin Reiche
#
# Contact: dominic@bioinf.uni-leipzig.de
#          kristin@bioinf.uni-leipzig.de
# http://www.bioinf.uni-leipzig.de/Software/NcDNAlign/
#
########################################################

use NcDNAlign;
use strict;
use Math::BigFloat;
use Getopt::Long;
use Pod::Usage;

no warnings 'recursion';

my ($opt_conf, $opt_out, $opt_test, $opt_help, $opt_version, $opt_man, $opt_maxDist, $opt_s) =
	('', 0, 0, '', '', '', 30, 0);

GetOptions(
	"c|conf=s" 		=> \$opt_conf,
	"d|dist:i" 		=> \$opt_maxDist,
	"o|out:i" 		=> \$opt_out,
	"t|test:i" 		=> \$opt_test,
	"h|help"		=> \$opt_help,
	"v|version"		=> \$opt_version,
	"man"			=> \$opt_man,
	"s|silent:i"	=> \$opt_s
) or pod2usage(2);

pod2usage(1) if ($opt_help || $opt_s !~ /^[1|0]$/);
pod2usage(-exitstatus => 0, -verbose => 2) if $opt_man;

if($opt_version) {
	print "This is ncDNAlign.3.mergeGwAln.pl, part of the NcDNAlign package, version: $VERSION.\n$url\n";
	exit;
}

if(!$opt_conf){
	pod2usage(1);
}

if($opt_out !~ /^[0|1]$/ || $opt_test !~ /^[0|1]$/) {
	pod2usage(1);
}

# ------ Start ----------
loadConfig($opt_conf);

# Further configuration
my $maxOverlap 	= 25; # Maximal overlap allowed to classify a specific type of pairwise alignments as consistent
my $blastDir 	= "$C{p_dir}/blast/"; # Where are the results?

opendir(DIR, $blastDir) or die $die_msg."Could not open directory '$blastDir'!";
my @Files = readdir(DIR);
close DIR;

my %hashData;
my %hashData_Info;
my $maxNeighbours=0;
my $file_merged;

main();

sub main {
	# Read blast results.
    foreach my $file(sort {$a cmp $b} @Files) {
		$file = $blastDir.$file;
		$file_merged = $file.".merged";

		unlink("$file_merged.cl.log") if (-e "$file_merged.cl.log");
		unlink("$file_merged.cl.tmp") if (-e "$file_merged.cl.tmp");
		unlink("$file_merged.cl.out") if (-e "$file_merged.cl.out");

		if($file !~ /^.*\.out$/ || $file =~ /cl.out/){ next; }

		open(IN, "$file") or die $die_msg."Could not open blast output '$file'!";
		print "file=$file\n" if($opt_s==0);

		while(<IN>) {
		    my $line = $_;
		    # example
		    # idPos_308583|scaffold_1 paired_scaffold_33      96.15   26      1       0       18      43      287046  287021  5e-04   44.1
		    # Prochlorococcus_marinus_str_MIT_9312:729516:2717:+      Anabaena_variabilis_ATCC_29413:72872:2760:+     75.46   436     107     0       1187    1622    1938    2373    1.1e-124        443.0
		    if($line =~ /^(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s*$/) {
				my $qID 			= $1;
				my $sID 			= $2;
				my $percIdent 		= $3;
				my $alignLength 	= $4;
				my $countMismatch 	= $5;
				my $countGapOpen 	= $6;
				my $qsID 			= 'query='.$qID.' subject='.$sID;
				my $qStart 			= $7;
				my $qEnd 			= $8;
				my $sStart 			= $9;
				my $sEnd 			= $10;
				my $eValue 			= $11;
				my $bitScore 		= $12;
				my $strand 			= 1;

				# Swap positions and change strand, if necessary
				if($sEnd < $sStart) {
			    	($sEnd, $sStart) = ($sStart, $sEnd);
			    	$strand = 0;
				}

				# Nearby hits should have same reading direction! Otherwise they do not belong to same alignment
				$qsID .= ' strand='.$strand;

				# $qsID has now query, subject and strand info

				my $infoKey = $qStart.'_'.$qEnd.'_'.$sStart.'_'.$sEnd;
				push @{$hashData{$qsID}}, $infoKey;

				$hashData_Info{$qsID.'_'.$infoKey} = $percIdent.' '.$alignLength.' '.$countMismatch.' '.$countGapOpen.' '.$eValue.' '.$bitScore;
		    }
		    elsif($line =~ /^\s*$/){ next; }
		    else {
			    die $die_msg."Wrong format in blast output '$file': line='$line'\n 1=$1, 2=$2, 3=$3, 4=$4, 5=$5, 6=$6, 7=$7, 8=$8, 9=$9, 10=$10, 11=$11, 12=$12!";
		    }
		}
		close(IN);

		unlink("$file_merged") if (-e $file_merged);

		# For each input file a separate output file is created
		merge_blast_alignments();
		# Test if all blast hits of file are found at least in one region of file_merged
		test_merged_alignments($file) if($opt_test==1);

		%hashData = ();
		%hashData_Info = ();

		unlink "$file_merged.cl.log";
		unlink "$file_merged.cl.tmp";
		unlink "$file_merged.cl.out";

    }# end foreach file
}# end of sub main


# PURPOSE: Merges overlapping HSPs.
# Takes care that nothing is merged which is inconsistent.
# For each possible set of consistent HSPs a union is created.
sub merge_blast_alignments {
	# for all blast hits on same query, subject and strand
    foreach my $k_qsID(keys %hashData) {
		# Built graph of neighbours. For each region look for all neighboured regions.
		my %neighGraph;
		for(my $i=0; $i<=@{$hashData{$k_qsID}}; $i++) {
	  		for(my $j=$i+1; $j<=@{$hashData{$k_qsID}}; $j++) {
	      		if($i==$j) { next; }

				if(exists $hashData{$k_qsID}->[$i] && exists $hashData{$k_qsID}->[$j]) {
		      		#print "i=$i '$hashData{$k_qsID}->[$i]', j=$j '$hashData{$k_qsID}->[$j]'\n";

		      		if(check_neighbour($hashData{$k_qsID}->[$i], $hashData{$k_qsID}->[$j]) == 1) {
			  			$neighGraph{$hashData{$k_qsID}->[$i]}{$hashData{$k_qsID}->[$j]} = 1;
			  			$neighGraph{$hashData{$k_qsID}->[$j]}{$hashData{$k_qsID}->[$i]} = 1;
		      		}
				}
	  		}
        }

		# All regions in @{$hashData{$k_qsID}}, but not found in neighGraph,
		# i.e. they do not have neighbours, must be taken separately.
		for (my $i=0; $i<=@{$hashData{$k_qsID}}; $i++) {
			if(exists $hashData{$k_qsID}->[$i]) {
		    	if(! exists $neighGraph{$hashData{$k_qsID}->[$i]} ){
		    		#print "(take_region1) hashData{$k_qsID}->[$i]=$hashData{$k_qsID}->[$i]\n";
		    		take_region($k_qsID, $hashData{$k_qsID}->[$i], 0, $hashData{$k_qsID}->[$i], "-");
		    	}
			}
		}

		# Resolve different set of neighbours = all connected components in graph --> single linkage
		my %hashNeighbours;    # Holds the results
		my %hashRememberGraph; # Which nodes were already seen?
		my $k = 1;

		foreach my $kg (keys %neighGraph) {
		    if(!exists $hashRememberGraph{$kg}) {
				push(@{$hashNeighbours{$k}}, $kg);
				$hashRememberGraph{$kg} = 1;
				next_round_single_linkage($kg, \%neighGraph, \%hashRememberGraph, \%hashNeighbours, $k);
				$k++;
		    }
		}

		# hashNeighbours contains now groups of regions which may be merged.
		# Need to check in each group if there occur inconsistencies
		foreach my $kn (keys %hashNeighbours) {
			# If current group has only one element take it and continue. Is actually obsolete!
		    if(@{$hashNeighbours{$kn}}==0) {
		    	#print "(take_region2) hashNeighbours{$kn}->[0]=$hashNeighbours{$kn}->[0]\n";
				take_region($k_qsID, $hashNeighbours{$kn}->[0], $file_merged, 0, $hashNeighbours{$kn}->[0], "-");
				die $die_msg."Only one member in current group! But this should not be the case!";
				next;
		    }

	    	# More than one element in group
	   	 	my %matrixInc;		# Which neighbours are inconsistent?
	    	my %matrixNeigh;    # Which nodes are neighbours?
	    	for(my $i=0; $i<=@{$hashNeighbours{$kn}}; $i++) {
				my $a = $hashNeighbours{$kn}->[$i];
				for(my $j=$i+1; $j<=@{$hashNeighbours{$kn}}; $j++) {
			    	if($i == $j){ next; }

			    	if(exists $hashNeighbours{$kn}->[$j]) {
				    	my $b = $hashNeighbours{$kn}->[$j];

					    # A group is a chain of neighboured regions,  but this doesn't mean
					    # that each region is a neighbour of all other regions.
					    # But there must be at least one neighbour to each region.
					    # For matrixNeigh do not insert edges between regions which are no neighbours.
					    if(check_neighbour($a, $b) == 0){
					    	next;
					    }

					    # Now I know 'a' and 'b' are direct neighbours --> check consistence
					    my ($inc, $stringCases) = check_consistence($a,$b);

			    		if($inc == 1) {
			    			$matrixNeigh{$a}{$b} = 1;
			    			$matrixNeigh{$b}{$a} = 1;
			    		}
			    		elsif($inc == 0) {
			    			$matrixInc{$a}{$b} = 1;
			    			$matrixInc{$b}{$a} = 1;
			    		}
			    		else{
			    			die $die_msg."ERROR: inc=$inc not known!";
			    		}
			    	}
				}
	    	}

	    	# All regions of matrixInc, but not found in matrixNeigh, must be taken separately.
	    	my @KeysMatrixInc = keys %matrixInc;
	    	for(my $i=0; $i<=$#KeysMatrixInc; $i++) {
				my $a = $KeysMatrixInc[$i];
				if(!exists $matrixNeigh{$a}){
					#print "(take_region3) a=$a\n";
					take_region($k_qsID, $a, 0, $a, "-");
				}
	    	}

            # Built consistence graph.
            # Two nodes x and y are linked if at least one path between x and y exists in 'matrixNeigh',
	    	# which does not contain nodes which are inconsistent to each other.
	    	my @KeysMatrix = keys %matrixNeigh;

	    	if($#KeysMatrix < 0){ next; }	# No consistent neighbours. Inconsistent regions were already taken --> continue
	   		my %matrixCons;             	# Two nodes are connected if there exists at least one consistent path
	    	my %hashRememberGraph;      	# Which nodes were already seen?
	    	my @currentPath;

	    	foreach my $kg (@KeysMatrix) {
				if(!exists $hashRememberGraph{$kg}) {
			    	$hashRememberGraph{$kg} = 1;
			    	push(@currentPath, $kg);
			    	next_round_built_consistence_graph($kg, \%matrixNeigh, \%matrixInc, \%hashRememberGraph, \%matrixCons, \@currentPath);
				    @currentPath = ();
				    %hashRememberGraph=();
				}
	    	}

			# Write consistence graph in DIMACS format (needed for cliquer.c by Sampo Niskanen and Patric Ostergard)
			my @Keys_a = keys %matrixCons;
			open(CL, "> $file_merged.cl.tmp") or die $die_msg."Could not create file '$file_merged.cl.tmp'!";
			print CL "p edge ", $#Keys_a+1," 0\n";
			# Get first a number code for each node
			my %hashId2Number;
			my %hashNumber2Id;
			my $cnt=0;
			my $what='';
			my $avgEval=0;
			my $sum=Math::BigFloat->new();
			my $sumReturn;

			foreach my $k_a(@Keys_a) {
			    $hashId2Number{$k_a}=++$cnt;
			    $hashNumber2Id{$cnt}=$k_a;

			    $what = join(",", @Keys_a); # comma separated list of used hit positions

			    my @e = split(/\s/, $hashData_Info{$k_qsID.'_'.$k_a});
			    my $x = Math::BigFloat->new($e[4]);
			    $sum->badd($x);
			}

			# calculate average eValue of merged hits
			$sum->bdiv(scalar(@Keys_a));
			$sum->bround(3);
			$sumReturn=$sum->bsstr();

		    foreach my $k_a(@Keys_a) {
		        my @Keys_b = keys %{$matrixCons{$k_a}};
				foreach my $k_b(@Keys_b) {
			    	print CL "e $hashId2Number{$k_a} $hashId2Number{$k_b}\n";
			    	delete $matrixCons{$k_a}{$k_b};
			    	delete $matrixCons{$k_b}{$k_a};
				}
		    }
		    close(CL);

		    ## Now get all maximal cliques from the consistence graph 'matrixCons' = sets of alignments which will be merged
		    ## Use the cliquer.c code by Sampo Niskanen, Patric Ostergard.
		    ## Usage: -a: Find all cliques;  -x: Requires cliques to be maximal;  -u: unweighted graph
		    ##        -m 1: Searching for all maximal cliques with size at least 1. Needed as otherwise cliquer would only look for maximum cliques,
		    ##              i.e. cliques which do not have a smaller number of nodes than the clique with the maximal number of nodes.
		    ##              '-u' says that all edges are weighted by 1, hence '-m 1' allows maximal cliques with weight at least 1.
		    ##        -1: Number vertices 1 to n instead of 0 to n-1
		    ##        -q: Quiet mode

		    my $sysCommand = "cl -a -x -u -m 1 -1 -q $file_merged.cl.tmp > $file_merged.cl.out 2> $file_merged.cl.log";

		    $maxNeighbours = @{$hashNeighbours{$kn}} if( @{$hashNeighbours{$kn}} > $maxNeighbours);

		    if($opt_out==1) {
				print "arrayNeighbours=@{$hashNeighbours{$kn}} ($maxNeighbours=$maxNeighbours)\n";
				print "sysCmd=$sysCommand\nwhat=$what\n\n"
		    }
		    system($sysCommand) == 0 or die $die_msg."Could not call '$sysCommand'";

		    # Read in cliques. Output format: size=%d, weight=%d:   %d ... %d
		    open(CL, "$file_merged.cl.out") or die $die_msg."Could not open file '$file_merged.cl.out'";
		    my @filedata = <CL>;
		    close(CL);

		    my %mergeCandidates; # Collects all regions which will be merged to one overall region
		    my $cntCand = 0;

		    foreach my $line(@filedata) {
				$cntCand++;

				if($line=~/^\s*size\S+\s+weight\S+(.+)$/) {
					my @aClique = ($1 =~ /(\d+)/g);
					foreach my $a(@aClique){ push(@{$mergeCandidates{$cntCand}}, $hashNumber2Id{$a}); }
				}
				elsif($line =~ /^\s+$/){ next; }
				else{ die $die_msg."while reading output of 'cl'. Line '$line' not known!";}
		    }

		    #foreach my $kc(sort {$a <=> $b} keys %mergeCandidates) {
			#	my @arrayCand = @{$mergeCandidates{$kc}};
		    #}

		    # Merge regions
		    my %mergedRegions = merge_regions(\%mergeCandidates);

		    foreach my $kc(sort {$a <=> $b} keys %mergedRegions) {
				foreach my $ac(sort {$a cmp $b} @{$mergedRegions{$kc}}) {
					#print "(take_region4) ac=$ac\n";
					take_region($k_qsID, $ac, 1, $what, $sumReturn);
				}
		    }
		} # end of foreach my $kn (keys %hashNeighbours)
    }# end of foreach k_qsID
} # end of sub merge_blast_alignments


# PURPOSE: Merge regions
sub merge_regions {
    my ($refMergeCandidates) = @_;

    my %hashMergedRegions;
    my $cnt=0;

    foreach my $k (keys %{$refMergeCandidates}) {
		# Get minimal and maximal positions of merged regions
		my @arrayRegions = @{$refMergeCandidates->{$k}};
		my $qStartMin = -1;
		my $qEndMax = 0;
		my $sStartMin = -1;
		my $sEndMax = 0;

		foreach my $region (@arrayRegions) {
			#print "(1) region=$region\n";
		    my ($qStart, $qEnd, $sStart, $sEnd) = parse_pos($region);

		    if($qStartMin==-1){ $qStartMin = $qStart; }
		    elsif($qStart < $qStartMin){ $qStartMin = $qStart; }

		    if($sStartMin==-1){ $sStartMin = $sStart; }
		    elsif($sStart < $sStartMin){ $sStartMin = $sStart; }

		    if($qEnd > $qEndMax){ $qEndMax = $qEnd; }

		    if($sEnd > $sEndMax){ $sEndMax = $sEnd; }
		}

		push(@{$hashMergedRegions{$cnt++}}, $qStartMin.'_'.$qEndMax.'_'.$sStartMin.'_'.$sEndMax);
    }

    return %hashMergedRegions;
}



# PURPOSE: Look if between two nodes a consistent path exists.
#          When a valid path is found all corresponding nodes are linked
#          in the consistence graph.
sub next_round_built_consistence_graph {
    my ($currentId, $refHashNeigh, $refHashInc, $refHashRemember, $refMatrixCons, $refCurrentPath) = @_;

    # Check if current path shows an inconsistence
    my $incFound=0;
    for(my $i=0; $i<=$#$refCurrentPath; $i++) {
		my $a = $refCurrentPath->[$i];
		for(my $j=$i+1; $j<=$#$refCurrentPath; $j++) {
	    	my $b = $refCurrentPath->[$j];

			if($refHashInc->{$a}{$b}) { # AVOID WARNING
				if($refHashInc->{$a}{$b} == 1 || $refHashInc->{$b}{$a} == 1) {  # In current path is an inconsistence
					$incFound=1; last;
				}
			}
		}
    }

    # Current path does not show any inconsistence.
    # Hence all nodes on this path can be pairwise linked in the consistency graph
    if($incFound==0) {
		for(my $i=0; $i<=$#$refCurrentPath; $i++) {
	    	my $a = $refCurrentPath->[$i];
	    	for(my $j=$i+1; $j<=$#$refCurrentPath; $j++) {
				my $b = $refCurrentPath->[$j];
				if(lc $a eq lc $b){ next; }
				$refMatrixCons->{$a}{$b}=1; $refMatrixCons->{$b}{$a}=1;
	    	}
		}
    }

    # Continue search
    foreach my $k_2 (keys %{$refHashNeigh->{$currentId}}) {
		if(!exists $refHashRemember->{$k_2} && $refHashNeigh->{$currentId}{$k_2} == 1) {
	    	$refHashRemember->{$k_2} = 1;
	    	# Continue path with new node
	    	push(@{$refCurrentPath}, $k_2);
	    	next_round_built_consistence_graph($k_2, $refHashNeigh, $refHashInc, $refHashRemember, $refMatrixCons, $refCurrentPath);
	    	pop(@{$refCurrentPath});
		}
    }
}# end of sub next_round_built_consistence_graph


# PURPOSE: True single linkage (depth first search) implementation
sub next_round_single_linkage {
    my ($currentId, $refHashData, $refHashRemember, $refHashClasses, $classCnt) = @_;

    foreach my $k_2 (keys %{$refHashData->{$currentId}}) {
        if(exists $refHashData->{$currentId}{$k_2}) {
            if(!exists $refHashRemember->{$k_2}) {
		push(@{$refHashClasses->{$classCnt}}, $k_2);
                $refHashRemember->{$k_2} = 1;
                next_round_single_linkage($k_2, $refHashData, $refHashRemember, $refHashClasses, $classCnt);
            }
        }
    }
}# end of sub next_round_single_linkage


# PURPOSE: Check consistence of two HSPs.
# Two HSPs are inconsistent if query regions are of different order than subject regions and vice versa.
# RETURNS: 1 if consistent
#          0 otherwise
sub check_consistence {
    my (@pos) = @_;
    my @sortedPos = my_sort(@pos);

    ## Sort positions
    my $a = $sortedPos[0];
    #print "a='$a'\n";
    my ($aqStart, $aqEnd, $asStart, $asEnd) = parse_pos($a);
    my $b = $sortedPos[1];
    #print "b='$b'\n";
    my ($bqStart, $bqEnd, $bsStart, $bsEnd) = parse_pos($b);

    my @cases=(-1,-1,-1,-1);

    ## Check, which lines cross each other: 'a' with 'c' or 'a' with 'd' or 'b' with 'c' or 'b' with 'd'
    ## a = aqStart - asStart; b = aqEnd - asEnd; c = bqStart - bsStart; d = bqEnd - bsEnd
    ## Two lines x and y cross each other if ...
    # a with c
    if($aqStart <= $bqStart && $asStart >= $bsStart || $aqStart >= $bqStart && $asStart <= $bsStart) {
    	$cases[0] = 1;
    }
    else{
    	$cases[0] = 0;
    }

    # a with d
    if($aqStart <= $bqEnd && $asStart >= $bsEnd || $aqStart >= $bqEnd && $asStart <= $bsEnd){
    	$cases[1] = 1;
    }
    else{
    	$cases[1] = 0;
    }

    # b with c
    if($aqEnd <= $bqStart && $asEnd >= $bsStart || $aqEnd >= $bqStart && $asEnd <= $bsStart){
    	$cases[2] = 1;
    }
    else{
    	$cases[2] = 0;
    }

    # b with d
    if($aqEnd <= $bqEnd && $asEnd >= $bsEnd || $aqEnd >= $bqEnd && $asEnd <= $bsEnd) {
    	$cases[3] = 1;
    }
    else{
    	$cases[3] = 0;
    }

    my $stringCases = '';
    foreach my $c(@cases){
    	$stringCases .= $c;
    }

    if($stringCases eq ''
       || $stringCases eq '1110' || $stringCases eq '0111' || $stringCases eq '0110'
       || $stringCases eq '0101' || $stringCases eq '0100' || $stringCases=~/\-1/) {
		die $die_msg."stringCases='$stringCases' is not possible! ".
	    " (aqStart=$aqStart, aqEnd=$aqEnd, asStart=$asStart, asEnd=$asEnd, bqStart=$bqStart, bqEnd=$bqEnd, bsStart=$bsStart, bsEnd=$bsEnd)";
    }
    elsif($stringCases eq '1111'){ return (0, $stringCases); }
    elsif($stringCases eq '1101'){ return (0, $stringCases); }
    elsif($stringCases eq '1100'){ return (0, $stringCases); }
    elsif($stringCases eq '1011'){ return (0, $stringCases); }
    elsif($stringCases eq '1010'){ return (0, $stringCases); }
    elsif($stringCases eq '1001'){ return (0, $stringCases); }
    elsif($stringCases eq '1000'){ return (0, $stringCases); }
    elsif($stringCases eq '0011'){ return (0, $stringCases); }
    elsif($stringCases eq '0010') {
		my $diff;
		if($bqStart <= $aqEnd){
			$diff = $aqEnd - $bqStart + 1;
		}
		elsif($bsStart <= $asEnd) {
			$diff = $asEnd - $bsStart + 1;
		}
		else{
			die $die_msg."stringCases='$stringCases': bqStart > aqEnd or bsStart > asEnd (bqStart=$bqStart, aqEnd=$aqEnd, bsStart=$bsStart, asEnd=$asEnd)\n\n";
		}

		if($diff < $maxOverlap) {
			return (1, $stringCases);
		}
		else{
			return (0, $stringCases);
		}
    }
    elsif($stringCases eq '0001') {
    	return (0, $stringCases);
    }
    elsif($stringCases eq '0000'){
    	return (1, $stringCases);
    }
    else{
    	die $die_msg."ERROR: Unknown case: stringCases = '$stringCases'";
    }
}# end of sub check_consistence


# PURPOSE: Checks if two regions are neighbours.
# They are neighbours if both query and subject regions have maximal distance of 'maxDiff'.
# RETURNS: 1 if same neighbourhood
#          0 otherwise
sub check_neighbour {
    my ($a, $b) = @_;
	#print "a='$a', b='$b'\n";
    my ($aqStart, $aqEnd, $asStart, $asEnd) = parse_pos($a);
    my ($bqStart, $bqEnd, $bsStart, $bsEnd) = parse_pos($b);

    my $qOverlap = overlapping($aqStart, $aqEnd, $bqStart, $bqEnd);
    my $sOverlap = overlapping($asStart, $asEnd, $bsStart, $bsEnd);
    my $qDiff = $opt_maxDist+1;
    if($aqEnd < $bqStart){ $qDiff = $bqStart-$aqEnd+1; }
    elsif($bqEnd < $aqStart){ $qDiff = $aqStart-$bqEnd+1;}
    my $sDiff = $opt_maxDist+1;
    if($asEnd < $bsStart){ $sDiff = $bsStart-$asEnd+1; }
    elsif($bsEnd < $asStart){ $sDiff = $asStart-$bsEnd+1;}

    # Do both overlap?
    if($qOverlap==1 && $sOverlap==1){ return 1; }
    elsif($qOverlap==1 && $sDiff<=$opt_maxDist){ return 1; }
    elsif($qDiff<=$opt_maxDist && $sOverlap==1){ return 1; }
    elsif($qDiff <= $opt_maxDist && $sDiff <= $opt_maxDist){ return 1; }
    else{ return 0; }
}# end of sub check_neighbour



sub my_sort {
    my (@array) = @_;

    my @sortedArray = sort { ($a=~/(\d+)/g)[0] <=> ($b=~/(\d+)/g)[0] } @array;;

    return @sortedArray;
}



sub parse_pos {
    my ($pos) = @_;

    # Parse Pos
    if($pos =~ /(\S+)_(\S+)_(\S+)_(\S+)/) {
		return ($1, $2, $3,  $4);
    }
    elsif($pos eq '') {
    	die $die_msg."ERROR: pos=\"\"";
    }
    else{
    	die $die_msg."ERROR: Could not parse position (pos=$pos)";
    }
}


# PURPOSE: Creating output: Takes all merged regions and prints them into a file
sub take_region {
	my ($qsID, $region, $combined, $what, $sumEval) = @_;

	open(OUT, ">>$file_merged") or die $die_msg."(take_region): Could not open file '$file_merged'";
	#print "region=$region\n";
	my $string = $qsID.' '.$region.' combined='.$combined;
	#print "string=$string\n";
	$string =~ /^query\=(\S+)\s+subject\=(\S+)\s+strand\=(\S+)\s+(\S+)\s+combined\=(\d)\s*$/ or die $die_msg."(take_region): Could not parse '$string'!";
	my $idPositions 			= $1;
	my $nameScaffold_subject 	= $2;
	my $strand 					= $3;
	my $posHit 					= $4;
	$combined 					= $5;
	my ($qStart, $qEnd, $subStart, $subEnd) = parse_pos($posHit);

	my $alignLength = $qEnd - $qStart + 1;

	# re-swap the positions
	# then sS > sE, but they were changed at the beginning
	if($strand == 0) {
	    ($subEnd, $subStart) = ($subStart, $subEnd);
	}

	if($combined==1) {
		my $string =	"combined=$combined query=$idPositions subject=$nameScaffold_subject alnLen=$alignLength ".
						"qS=$qStart qE=$qEnd sS=$subStart sE=$subEnd eV=$sumEval how=$what\n\n";
		print $string if($opt_out==1);
		print OUT $string;
	}
	else {
		my $string_Info = $hashData_Info{$qsID.'_'.$posHit};
		$string_Info =~ /^\s*(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s*/ or die $die_msg."(take_region): Could not parse string_Info '$string_Info'!";

		my $percIdent 		= $1;
		$alignLength 		= $2;
		my $countMismatch 	= $3;
		my $countGapOpen 	= $4;
		my $eValue 			= $5;
		my $bitScore 		= $6;

		my $string =	"combined=$combined query=$idPositions subject=$nameScaffold_subject percIdent=$percIdent ".
						"alnLen=$alignLength qtMM=$countMismatch qtGO=$countGapOpen ".
						"qS=$qStart qE=$qEnd sS=$subStart sE=$subEnd eV=$eValue bitS=$bitScore how=$what\n\n";
		print $string if($opt_out==1);
		print OUT $string;
	}

	close(OUT);
}# end of sub take_region



# PURPOSE: Test if all blast hits of file are found at least in one region of file_merged
sub test_merged_alignments
{
    my ($file_blast) = @_;

    my %hashBlast;
    my %hashMerged;

    open(IN, "$file_blast") or die $die_msg."(test_merged_alignments): Could not open file '$file_blast'";
    my @filedata_blast = <IN>;
    close(IN);

    open(IN, "$file_merged") or die $die_msg."(test_merged_alignments): Could not open file '$file_merged'";
    my @filedata_merged = <IN>;
    close(IN);

    ## Read blast
    foreach my $line_blast (@filedata_blast) {
		if($line_blast =~ /\s*^(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s*$/) {
			my $qID 	= $1;
			my $sID 	= $2;
			my $qsID 	= $qID.'_'.$sID;
			my $qStart 	= $7;
			my $qEnd 	= $8;
			my $sStart 	= $9;
			my $sEnd 	= $10;
			my $eValue 	= $11;

			if($eValue<=$C{cutoff_evalue}) {
				#print "eValue=$eValue accepted\n";
			}
			else {
				#print "eValue=$eValue declined\n";
				next;
			}

			# Swap positions, if necessary
			($qEnd, $qStart) = ($qStart, $qEnd) if($qEnd < $qStart);

			($sEnd, $sStart) = ($sStart, $sEnd) if($sEnd < $sStart);

			push(@{$hashBlast{$qID.'_'.$sID}}, $qStart.'_'.$qEnd.'_'.$sStart.'_'.$sEnd);
		}
		elsif($line_blast =~ /^\s*$/){ next; }
		else{ die $die_msg."(test_merged_alignments): Wrong format in blast output '$file_blast':\n line='$line_blast'($file_blast, $file_merged) !";}
    }

    ## Read merged

    foreach my $line_merged (@filedata_merged) {
		#combined=0 query=Bacillus_subtilis.gbk:30035:5493:+ subject=Bacillus_cereus:1:5224283:+ percIdent=98.65 alnLen=74 qtMM=1 qtGO=0 qS=1984 qE=2057 sS=4650152 sE=4650079 eV=9.0e-35 bitS=145.0 1984_2057_4650079_4650152
		#combined=1 query=Bacillus_subtilis.gbk:633411:6797:+ subject=Bacillus_cereus:1:5224283:+ alnLen=1554 qS=1571 qE=3124 sS=145493 sE=147048 1665_3124_145589_147048,1571_1641_145493_145563
		if($line_merged =~ /query\=(\S+).+subject\=(\S+).+qS\=(\d+).+qE\=(\d+).+sS\=(\d+).+sE\=(\d+)/) {
			my $q = $1;
			my $s = $2;
			my $sS= $5;
			my $sE= $6;
			# swap negative strand
			if($5 > $6){ $sS = $6; $sE = $5;}
			my $where = $3."_".$4."_".$sS."_".$sE;
			push(@{$hashMerged{$q.'_'.$s}}, $where);
		}
		elsif($line_merged =~ /^\s*$/){ next; }
		else{ die $die_msg."(test_merged_alignments): Wrong format in merged output '$file_merged':\n line='$line_merged' ($file_blast, $file_merged)!";}
	}

    # Check now if each blast hit is found at least once in filedata_merged
    my @KeysBlast = keys %hashBlast;
    my @KeysMerged = keys %hashMerged;
    # Same keys in merged and blast files?
    foreach my $kb (@KeysBlast) {
		if(!exists $hashMerged{$kb}){ die $die_msg."(test_merged_alignments): Key '$kb' of hashBlast not found in hashMerged ($file_blast, $file_merged)!"; }
    }

    foreach my $km (@KeysMerged) {
		if(!exists $hashBlast{$km}){ die $die_msg."(test_merged_alignments): Key '$km' of hashMerged not found in hashBlast ($file_blast, $file_merged)!"; }
    }

    # Check blast hits
    foreach my $kb (@KeysBlast) {

		my $found = 0;
		foreach my $posBlast (@{$hashBlast{$kb}}) {
		    my ($qStart, $qEnd, $sStart, $sEnd) = parse_pos($posBlast);

		    foreach my $posMerged (@{$hashMerged{$kb}}) {
				my ($qs, $qe, $ss, $se) = parse_pos($posMerged);
				# Corresponding merged region found?
				if($qStart >= $qs && $qEnd <= $qe && $sStart >= $ss && $sEnd <= $se) {
				    $found = 1;
				    last;
				}
		    }
	    	if($found == 0){
	    		die $die_msg."(test_merged_alignments): Blast hit '$kb: $posBlast' not found in merged regions '$file_merged' ($file_blast, $file_merged)!";
	    	}
		}
    }
}# end of sub test_merged_alignments

__END__

=head1 NAME

C<ncDNAlign.3.mergeGwAln.pl> - part of the NcDNAlign alignment pipeline, step (3)

Filtering and merging of BLAST hits (required blast input format: tabular output by -m 8). Ensures that all compatible hits in same genomic region (distance <= --dist) are merged to span complete genes.

=head1 SYNOPSIS

ncDNAlign.3.mergeGwAln.pl [options]

Type 

    $ ncDNAlign.3.mergeGwAln.pl --help

or

    $ ncDNAlign.3.mergeGwAln.pl --man 

for a detailed man page


=head1 OPTIONS

=over 8

=item B<-c, --conf> FILE

Path to central NcDNAlign configuration file [REQUIRED]

=item B<-d, --dist> INTEGER

Maximal nucleotide distance between BLAST HSPs to be considered as neighboured and, thus, potentially merged. Default: 30.

=item B<-o, --out> [0|1]

Printing detailed results at STDOUT ON(1) or OFF(0). Default: 0.

=item B<-t, --test> [0|1]

Should the results be verified? Enable consistency checks, e.g. ensure that each inputted HSP is also outputted (1=YES, 0=NO). Default: 0.

=item B<-s, --silent> [0|1]

Silent mode, avoid printing to STDOUT
0=OFF (Default), 1=ON

=item B<-v, --version>

Prints version information and exits.

=item B<-h, --help>

Prints a short help message and exits.

=item B<--man>

Prints a detailed manual page and exits.

=back

=head1 DESCRIPTION

Prevalently, sequences are less conserved in regions without base pair interactions, which might prevent the aligning procedures in getGenomwideAln.pl from extending the sequence alignment into such regions. Due to rearrangement, deletion, and duplication events during evolution, not all single local alignments lead to consistent global alignments. Therefore, compatible adjacent BLAST HSPs with a maximal distance of 30 nt are combined to ensure that heuristic global alignments span complete genes. Only appropriate HSPs exceeding a certain EVALUE-cutoff are reported and make it into a consecutive merging process. This merging algorithm is about calculating consistence graphs. Their maximal cliques define valid sets of combinable alignments. Output is written to *.merged files. They contain the merged BLAST hits.

=head1 EXAMPLES

$ ncDNAlign.3.mergeGwAln.pl -c config-file.cfg

=head1 AUTHORS

=over 8

=item Dominic Rose   (dominic@bioinf.uni-leipzig.de)

=item Kristin Reiche (kristin@bioinf.uni-leipzig.de)

=back

=head1 AVAILABILITY

http://www.bioinf.uni-leipzig.de/Software/NcDNAlign/

=cut


